# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

import time
from datetime import datetime
import uuid
import redis
import aloharedis
assert aloharedis.__version__ == '0.0.3'
from aloharedis.stores import *

from nose.tools import *

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! IMPORTANT: We use redis db=7 for these tests   !!!
# !!! and will DELETE ALL DATA from it after tested. !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DB = 7

def setup_module():
    global redis_client
    redis_client = redis.StrictRedis(db=DB)


def teardown_module():
    redis_client.flushdb()


def test_store():
    store = Store(redis_client)
    eq_(None, store.get('a'))
    ok_(store.set('a', 100))
    eq_('100', store.get('a'))
    ok_(store.delete('a'))
    ok_(not store.exists('a'))


def test_pickle_store():
    store = PickleStore(redis_client)
    data= uuid.uuid4()
    ok_(store.set('a', data))
    eq_(data, store.get('a'))
    eq_(data, store.getset('a', 200))
    eq_(200, store.get('a'))
    ok_(store.mset(dict(x=1, y=2, z=3)))
    eq_([1,2,3], store.mget(['x', 'y', 'z']))
    eq_([1,2,3], store.mget('x', 'y', 'z'))
    ok_(store.msetnx(dict(w=4, u=5, v=6)))
    ok_(not store.msetnx(dict(a=1, i=7, j=8, k=9)))
    ok_(store.setnx('b', 2))
    ok_(not store.setnx('b', 3))
    ok_(store.setex('c', 1, 'x'))
    ok_(store.exists('c'))
    time.sleep(2.0)
    ok_(not store.exists('c'))

    
def test_json_store():
    store = JsonStore(redis_client)
    data = [1,2,3]
    ok_(store.set('a', data))
    eq_(data, store.get('a'))
    with assert_raises(TypeError):
        store.set('a', datetime.now())

    
def test_mongo_extended_json_store():
    store = MongoExtendedJsonStore(redis_client)
    data= uuid.uuid4()
    ok_(store.set('a', data))
    eq_(data, store.get('a'))
    moment = datetime.now()
    store.set('a', moment)
    new = store.get('a').replace(tzinfo=None)
    eq_(moment.replace(microsecond=0), new.replace(microsecond=0))



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license
"""
Caches
======

"""

__all__ = ['CacheManager']

import time
from .stores import PickleStore

MAX_EXPIRE_SECONDS = 946656000  # timestamp of Jan. 1st, 2000
                                # objects' __expired__ large than this
                                # will be treated as a timestamp,
                                # otherwise it will be treated as seconds

class CacheManager(PickleStore):
    """Baseclass for the cache systems.  All the cache systems implement this
    API or a superset of it.

    :param default_timeout: the default timeout that is used if no timeout is
                            specified on :meth:`set`.
    """

    def __init__(self, store, key_prefix='<CACHE>', default_timeout=300):
        PickleStore.__init__(self, store)
        if key_prefix:
            self.key_prefix = str(key_prefix)
        else:
            self.key_prefix = ''
        self.default_timeout = default_timeout

    def get(self, key):
        """Looks up key in the cache and returns the value for it.
        If the key does not exist `None` is returned instead.

        :param key: the key to be looked up.
        """
        return PickleStore.get(self, self.key_prefix+key)

    def delete(self, key):
        """Deletes `key` from the cache.  If it does not exist in the cache
        nothing happens.

        :param key: the key to delete.
        """
        self.store.delete(self.key_prefix+key)

    def get_many(self, *keys):
        """Returns a list of values for the given keys.
        For each key a item in the list is created.  Example::

            foo, bar = cache.get_many("foo", "bar")

        If a key can't be looked up `None` is returned for that key
        instead.

        :param keys: The function accepts multiple keys as positional
                     arguments.
        """
        if self.key_prefix:
            keys = [self.key_prefix + key for key in keys]
        return self.mget(*keys)

    def get_dict(self, *keys):
        """Works like :meth:`get_many` but returns a dict::

            d = cache.get_dict("foo", "bar")
            foo = d["foo"]
            bar = d["bar"]

        :param keys: The function accepts multiple keys as positional
                     arguments.
        """
        values = self.get_many(*keys)
        result = dict()
        for i in range(len(keys)):
            result[keys[i]] = values[i]
        return result

    def set(self, key, value, timeout=None):
        """Adds a new key/value to the cache (overwrites value, if key already
        exists in the cache).

        :param key: the key to set
        :param value: the value for the key
        :param timeout: the cache timeout for the key (if not specified,
                        it uses the default timeout).
        """
        timeout = timeout or self.default_timeout
        if timeout>MAX_EXPIRE_SECONDS:
            timeout = max(0, timeout-int(time.time()))
        if self._dialect=='Redis':
            self.setex(self.key_prefix+key, value, timeout)
        else:
            self.setex(self.key_prefix+key, timeout, value)

    def add(self, key, value, timeout=None):
        """Works like :meth:`set` but does not overwrite the values of already
        existing keys.

        :param key: the key to set
        :param value: the value for the key
        :param timeout: the cache timeout for the key or the default
                        timeout if not specified.
        """
        timeout = timeout or self.default_timeout
        if timeout>MAX_EXPIRE_SECONDS:
            timeout = max(0, timeout-int(time.time()))
        success = self.setnx(self.key_prefix+key, value)
        if success:
            self.store.expire(self.key_prefix+key, timeout)

    def set_many(self, mapping, timeout=None):
        """Sets multiple keys and values from a mapping.

        :param mapping: a mapping with the keys/values to set.
        :param timeout: the cache timeout for the key (if not specified,
                        it uses the default timeout).
        """
        timeout = timeout or self.default_timeout
        if timeout>MAX_EXPIRE_SECONDS:
            timeout = max(0, timeout-int(time.time()))
        with self.store.pipeline() as pipe:
            for key in mapping:
                if self._dialect=='Redis':
                    pipe.setex(self.key_prefix+key, mapping[key], timeout)
                else:
                    pipe.setex(self.key_prefix+key, timeout, mapping[key])
            pipe.execute()
            
    def delete_many(self, *keys):
        """Deletes multiple keys at once.

        :param keys: The function accepts multiple keys as positional
                     arguments.
        """
        for key in keys:
            self.delete(key)

    def keys(self):
        return self.store.keys(self.key_prefix+'*')

    def clear(self):
        """Clears the cache.  Keep in mind that not all caches support
        completely clearing the cache.
        """
        if not self.key_prefix:
            return
        keys = self.keys()
        if keys:
            self.store.delete(*keys)

    def inc(self, key, delta=1):
        """Increments the value of a key by `delta`.  If the key does
        not yet exist it is initialized with `delta`.

        For supporting caches this is an atomic operation.

        :param key: the key to increment.
        :param delta: the delta to add.
        """
        return self.store.incr(self.key_prefix + key, delta)

    def dec(self, key, delta=1):
        """Decrements the value of a key by `delta`.  If the key does
        not yet exist it is initialized with `-delta`.

        For supporting caches this is an atomic operation.

        :param key: the key to increment.
        :param delta: the delta to subtract.
        """
        return self.store.decr(self.key_prefix + key, delta)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

"""
Serial Generators
=================

Example
-------

Below is a simple example to show how this module will work::

    >>> import redis
    >>> from aloharedis.serials import *
    >>> store = SerialsManager(redis.Redis(), '<MySerials>',
    ...     Serial('session.id', pattern=u'SID:%032x'),
    ...     PeriodSerial('sale-order', PERIOD_MONTHLY,
                         pattern=u'SO-%(moment)s-%(mumber)06d',
                         moment_format='%Y-%m')
    ... )
    >>> store.incr('session.id')
    u'SID:00000000000000000000000000000001'
    >>> store.incr('sale-order')
    u'SO-2012-06-000001'

"""

__all__ = ['SerialsManager', 'Serial', 'PeriodSerial',
           'PERIOD_YEARLY', 'PERIOD_MONTHLY', 'PERIOD_WEEKLY_SUNDAY_FIRST',
           'PERIOD_WEEKLY_MONDAY_FIRST', 'PERIOD_DAILY', 'PERIOD_HOURLY']

from datetime import datetime
from .stores import Store

PERIOD_YEARLY = '%Y'
PERIOD_MONTHLY = '%Y%m'
PERIOD_WEEKLY_SUNDAY_FIRST = '%Y%U'
PERIOD_WEEKLY_MONDAY_FIRST = '%Y%W'
PERIOD_DAILY = '%Y%m%d'
PERIOD_HOURLY = '%Y%m%d%H'


class Serial(object):

    def __init__(self, name, initial=0, increment=1, pattern=u'%d'):
        self.name = str(name)
        self.initial = int(initial)
        self.increment = int(increment)
        self.pattern = unicode(pattern)
        self._last_field = None

    def format(self, data):
        return self.pattern % data
    
    def next(self, store, key, increment=0):
        return self.format(self._increase(store, key, increment))
    
    def outdated(self, field):
        field = str(field)
        if field==self.name:
            return False # not outdated
        return None      # unknown

    def _increase(self, store, key, increment=0):
        increment = int(increment or self.increment)
        field = self.name
        if self.initial and self._last_field != field:
            store.hsetnx(key, field, self.initial)
        self._last_field = field
        number = store.hincrby(key, field, increment)
        return number


class PeriodSerial(Serial):

    def __init__(self, name, period, initial=0, increment=1,
                 pattern=u'%(moment)s%(number)06d', moment_format=None):
        Serial.__init__(self, name, initial, increment, pattern)
        self.period = str(period)
        self._last_period = None
        if moment_format!=None:
            self.moment_format = unicode(moment_format)
        else:
            self.moment_format = unicode(self.period)
    
    def _increase(self, store, key, increment=0):
        increment = int(increment or self.increment)
        moment = datetime.now()
        field = '{0}:{1}'.format(self.name, moment.strftime(self.period))
        if self.initial and self._last_field != field:
            store.hsetnx(key, field, self.initial)
        self._last_field = field
        number = store.hincrby(key, field, increment)
        return dict(moment=moment.strftime(self.moment_format), number=number)

    def outdated(self, field):
        field = str(field)
        if not field.startswith(self.name+':'):
            return None  # None means unknown
        moment = datetime.now()
        current = '{0}:{1}'.format(self.name, moment.strftime(self.period))
        if field<current:
            return True  # outdated
        else:
            return False # not outdated


class SerialsManager(Store):
    
    def __init__(self, store, key='<SERIALS>', *serials):
        Store.__init__(self, store)
        self._key = str(key)
        self.store.hsetnx(self.key, '<type>', self.__class__.__name__)
        self._serials = dict()
        for serial in serials:
            self.add_serial(serial)

    @property
    def key(self):
        return self._key
    
    def add_serial(self, serial):
        name = str(serial.name)
        if name in self._serials:
            #raise RuntimeError('serial "%s" already registered' % name)
            return False
        self._serials[name] = serial
        return True
    
    def has_serial(self, serial):
        name = str(getattr(serial, 'name', serial))
        return name in self._serials
    
    def remove_serial(self, serial):
        name = str(getattr(serial, 'name', serial))
        return self._serials.pop(name, None)

    def incr(self, name, increment=0):
        name = str(name)
        serial = self._serials[name]
        result = serial.next(self.store, self.key, increment)
        return result
    
    @property
    def keys(self):
        keys = set(self.store.hkeys(self.key))
        if '<type>' in keys:
            keys.remove('<type>')
        return keys

    def clean(self):
        fields = self.keys
        outdateds = set()
        for field in fields:
            for name in self._serials:
                serial = self._serials[name]
                status = serial.outdated(field)
                if status != None:
                    if status:
                        outdateds.add(field)
                    # either True or False means
                    # the serial knows the field
                    break
        for outdated in outdateds:
            self.store.hdel(self.key, outdated)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license
"""
Locks
=====

"""

__all__ = ['Lock', 'LockManager']

import time
from redis import WatchError
from redis.client import Lock as _Lock
import redis
from .stores import Store


class Lock(_Lock):
    # By using redis.client.Lock directly, the unit tests faild sometimes
    # Our solution is using the Redis' transaction to acquire the lock
    
    WAIT_FOREVER = float(2**31+1) # 1 past max unix time

    def acquire(self, blocking=True, max_wait=None):
        """
        Use Redis to hold a shared, distributed lock named ``name``.
        Returns True once the lock is acquired.

        If ``blocking`` is False, always return immediately. If the lock
        was acquired, return True, otherwise return False.
        """
        started = int(time.time())
        if max_wait:
            deadline = started + float(max_wait)
        else:
            deadline = self.WAIT_FOREVER
        sleep = self.sleep
        timeout = self.timeout
        while True:
            got = False
            with self.redis.pipeline() as trans:
                unixtime = int(time.time())
                if unixtime>deadline:
                    return False
                if timeout:
                    timeout_at = unixtime + timeout
                else:
                    timeout_at = Lock.LOCK_FOREVER
                timeout_at = float(timeout_at)
                try:
                    trans.watch(self.name)
                    exists = trans.exists(self.name)
                    trans.multi()
                    if not exists:
                        if timeout:
                            if self._dialect=='Redis':
                                trans.setex(self.name, timeout_at, timeout)
                            else:
                                trans.setex(self.name, timeout, timeout_at)
                        else:
                            trans.set(self.name, timeout_at)
                        trans.execute()
                        self.acquired_until = timeout_at
                        got = True
                except WatchError:
                    pass
            if got:
                return True
            if not blocking:
                return False
            time.sleep(sleep)
    
    @property
    def _dialect(self):
        if isinstance(self.redis, redis.Redis):
            return 'Redis'
        return ''


class LockManager(Store):
    
    def __init__(self, store, key_prefix='<LOCK>',
                 default_timeout=10, default_sleep=0.1):
        Store.__init__(self, store)
        if key_prefix:
            self.key_prefix = str(key_prefix)
        else:
            self.key_prefix = ''
        self.default_timeout = default_timeout
        self.default_sleep = default_sleep
        self._locks = dict()

    def configure(self, name, timeout=None, sleep=None):
        name = str(name)
        timeout = timeout or self.default_timeout
        sleep = sleep or self.default_sleep
        lock = Lock(self.store, self.key_prefix+name, timeout, sleep)
        self._locks[name] = lock
        return lock
    
    def acquire(self, name, blocking=True):
        name = str(name)
        if name not in self._locks:
            self.configure(name)
        return self[name].acquire(blocking)
    
    def release(self, name):
        name = str(name)
        if name not in self._locks:
            self.configure(name)
        self[name].release()
    
    def __getitem__(self, name):
        name = str(name)
        if name not in self._locks:
            self.configure(name)
        return self._locks[name]
    
    def keys(self):
        return self.store.keys(self.key_prefix+'*')
    
    def force_release(self, name):
        name = str(name)
        self.delete(self.key_prefix+name)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

"""
Store
=====


"""

__all__ = ['Redis', 'Store', 'SerializationStore',
           'PickleStore', 'JsonStore', 'MongoExtendedJsonStore']

try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    import simplejson as json
except:
    import json
try:
    from bson import json_util
except:
    json_util = None
from redis import Redis
from redis import StrictRedis


class Store(object):
    
    def __init__(self, store):
        if not isinstance(store, StrictRedis):
            raise TypeError('"store" should an instance of redis.StrictRedis')
        if isinstance(store, Redis):
            self._dialect = 'Redis'
        else:
            self._dialect = 'StrictRedis'
        self._store = store
    
    @property
    def store(self):
        return self._store

    def __getattr__(self, name):
        return getattr(self.store, name)


class SerializationStore(Store):
    
    def serialize(self, value):
        # keep integers unserialized so incr() works
        if type(value) in [int, long]:
            return str(value)
        else:
            return '$' + self._serialize(value)
    
    def unserialize(self, value):
        if value!=None:
            if value.startswith('$'):
                return self._unserialize(value[1:])
            else:
                return int(value)

    def _serialize(self, value):
        raise NotImplementedError
    
    def _unserialize(self, value):
        raise NotImplementedError

    def get(self, key):
        value = self.store.get(key)
        return self.unserialize(value)

    def getset(self, key, value):
        value = self.serialize(value)
        value = self.store.getset(key, value)
        return self.unserialize(value)
    
    def mget(self, keys, *args):
        values = self.store.mget(keys, *args)
        for i in range(len(values)):
            values[i] = self.unserialize(values[i])
        return values

    def mset(self, mapping):
        data = dict()
        for key in mapping:
            data[key] = self.serialize(mapping[key])
        return self.store.mset(data)

    def msetnx(self, mapping):
        data = dict()
        for key in mapping:
            data[key] = self.serialize(mapping[key])
        return self.store.msetnx(data)

    def set(self, key, value):
        value = self.serialize(value)
        return self.store.set(key, value)
    
    def setex(self, key, seconds, value):
        if self._dialect=='Redis':
            seconds, value = value, seconds
        value = self.serialize(value)
        if self._dialect=='Redis':
            return self.store.setex(key, value, seconds)
        else:
            return self.store.setex(key, seconds, value)
    
    def setnx(self, key, value):
        value = self.serialize(value)
        return self.store.setnx(key, value)


class PickleStore(SerializationStore):
    
    def _serialize(self, value):
        return pickle.dumps(value)
    
    def _unserialize(self, value):
        return pickle.loads(value)


class JsonStore(SerializationStore):

    def _serialize(self, value):
        return json.dumps(value)
    
    def _unserialize(self, value):
        return json.loads(value)


class MongoExtendedJsonStore(SerializationStore):
    """
    http://www.mongodb.org/display/DOCS/Mongo+Extended+JSON
    """

    def _serialize(self, value):
        return json.dumps(value, default=json_util.default)
    
    def _unserialize(self, value):
        return json.loads(value, object_hook=json_util.object_hook)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

import time
from threading import Thread
import redis
import aloharedis
assert aloharedis.__version__ == '0.0.3'
from aloharedis.locks import *

from nose.tools import *

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! IMPORTANT: We use redis db=7 for these tests   !!!
# !!! and will DELETE ALL DATA from it after tested. !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DB = 7

def setup_module():
    global redis_client
    redis_client = redis.StrictRedis(db=DB)
    redis_client.flushdb()


def teardown_module():
    redis_client.flushdb()

monitor = list()

def run_block(num, locks):
    lock = locks['lock']
    got = lock.acquire(True)
    start = time.time()
    time.sleep(0.1)
    end = time.time()
    if got:
        monitor.append((num, start, end))
        lock.release()

def run_nonblock(num, locks):
    lock = locks['lock']
    got = lock.acquire(False)
    start = time.time()
    time.sleep(0.5)
    end = time.time()
    if got:
        monitor.append((num, start, end))
        lock.release()

def run_block_with_wait(num, locks):
    lock = locks['lock']
    got = lock.acquire(True, 0.5)
    start = time.time()
    time.sleep(0.1)
    end = time.time()
    if got:
        monitor.append((num, start, end))
        lock.release()

def test1():
    locks = LockManager(redis_client)
    with locks['a']:
        eq_(['<LOCK>a'], locks.keys())
    eq_([], locks.keys())

def test2():
    global monitor
    locks = LockManager(redis_client)
    locks.configure('lock')
    thlist = list()
    num = 10
    thlist = list()
    monitor = list()
    for i in range(num):
        t = Thread(target=run_block, args=(i+1, locks))
        t.start()
        thlist.append(t)
    for t in thlist:
        t.join()
    eq_(num, len(monitor))
    for i in range(num-1):
        p = monitor[i]
        n = monitor[i+1]
        ok_(n[1]>p[2])
    thlist = list()
    monitor = list()
    for i in range(num):
        t = Thread(target=run_nonblock, args=(i+1, locks))
        t.start()
        thlist.append(t)
    for t in thlist:
        t.join()
    eq_(1, len(monitor))
    thlist = list()
    monitor = list()
    for i in range(num):
        t = Thread(target=run_block_with_wait, args=(i+1, locks))
        t.start()
        thlist.append(t)
    for t in thlist:
        t.join()
    ok_(num>len(monitor))
    ok_(1<len(monitor))





# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

import time
from datetime import datetime
from random import random
import redis
import aloharedis
assert aloharedis.__version__ == '0.0.3'
from aloharedis.ohm import *

from nose.tools import *

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! IMPORTANT: We use redis db=7 for these tests   !!!
# !!! and will DELETE ALL DATA from it after tested. !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DB = 7

def setup_module():
    global redis_client
    redis_client = redis.StrictRedis(db=DB)


def teardown_module():
    redis_client.flushdb()


class User(Hash):
    username = StringAttr(as_key=True, min_length=3)
    name = UnicodeAttr()
    enabled = BooleanAttr(required=True)
    age = IntAttr()
    random = FloatAttr()

class Customer(User):
    __fixed_keys__ = True
    __ensure_new__ = True
    customer_id = StringAttr(as_key=True, min_length=8)


class ChatLog(Hash):
    __key_sep__ = '|'
    __key_prefix__ = 'chat'
    ts = TimestampAttr()
    username = StringAttr()
    msg = UnicodeAttr()
    def __init__(self, username='', msg=u''):
        self.username = username
        self.msg = msg

class Order(Hash):
    oid = PeriodSerialAttr(as_key=True, pattern='O-%(moment)s-%(number)04d')
    data = Attr()


def test_simple():
    mgr = Manager(redis_client)
    u1 = User()
    u1.username = 'guest'
    u1.name = 'Guest'
    u1.enabled = 'no'
    u1.random = random()
    ok_(type(u1.name)==unicode)
    ok_(type(u1.enabled)==bool)
    ok_(type(u1.age)==int)
    ok_(mgr.save(u1))
    users = mgr.find(User, 'guest')
    ok_('User:guest' in users)
    users = mgr.find(User)
    ok_('User:guest' in users)
    users = mgr.find(User, username='guest')
    ok_('User:guest' in users)
    u = mgr.load(User, users[0])
    eq_(u, u1)
    # __fixed_keys__ is False
    u.username = 'guest1'


def test_inherited():
    mgr = Manager(redis_client)
    u2 = Customer()
    u2.username = 'customer1'
    u2.name = 'Tom Smith'
    u2.enabled = 'yes'
    u2.customer_id = '1234567'
    with assert_raises(ValueError):
        mgr.save(u2)
    u2.customer_id = '12345678'
    mgr.save(u2)
    customers = mgr.find(Customer)
    ok_('Customer:customer_id:12345678:username:customer1' in customers)
    customers = mgr.find(Customer, customer_id=12345678)
    u = mgr.load(Customer, customers[0])
    eq_(u, u2)
    # __fixed_keys__ is True
    with assert_raises(RuntimeError):
        u.username = 'customer2'
    with assert_raises(RuntimeError):
        u2.customer_id = '87654321'
    # __ensure_new is True
    u3 = Customer()
    u3.username = 'customer1'
    u3.name = 'Jim Smith'
    u3.enabled = 'no'
    u3.customer_id = '12345678'
    with assert_raises(RuntimeError):
        mgr.save(u3)

def test_customize_key_serial_timestamp():
    ts = int(time.time())
    mgr = Manager(redis_client)
    log1 = ChatLog('tom', 'Hello! How are you?')
    log2 = ChatLog('joe', 'Fine, thank you! And you?')
    log3 = ChatLog('tom', 'I''m fine, too. Thank you!')
    ok_(mgr.save(log1))
    ok_(mgr.save(log2))
    ok_(mgr.save(log3))
    eq_(1, log1._id)
    eq_(2, log2._id)
    eq_(3, log3._id)
    eq_(log1, mgr.load(ChatLog, 'chat|1'))
    eq_(log2, mgr.load(ChatLog, 'chat|2'))
    eq_(log3, mgr.load(ChatLog, 'chat|3'))
    ok_(abs(log1.ts-ts)<=1)


def test_period_serial():
    mgr = Manager(redis_client)
    o1 = Order()
    eq_(None, o1.oid)
    o1.attach(mgr)
    o1.save()
    ok_(o1.oid.endswith('0001'))
    o2 = Order()
    mgr.save(o2)
    ok_(o2.oid.endswith('0002'))



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

__version__ = '0.0.3'

from stores import StrictRedis, Redis, Store, SerializationStore,\
                   PickleStore, JsonStore, MongoExtendedJsonStore
from caches import CacheManager
from locks import Lock, LockManager
from serials import Serial, PeriodSerial, SerialsManager,\
                    PERIOD_YEARLY, PERIOD_MONTHLY, PERIOD_WEEKLY_SUNDAY_FIRST,\
                    PERIOD_WEEKLY_MONDAY_FIRST, PERIOD_DAILY, PERIOD_HOURLY



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

"""
Object Hash Mapper
==================


"""

__all__ = [
    'STATE_NEW', 'STATE_DIRTY', 'STATE_UNCHANGED', 'MAX_EXPIRE_SECONDS',
    'Attr', 'EnumAttr', 'NumberAttr', 'IntAttr', 'FloatAttr',
    'TimestampAttr', 'SerialAttr', 'BooleanAttr', 'StringAttr',
    'PeriodSerialAttr', 'UnicodeAttr',
    'Hash', 'Manager'
]

import time
try:
    import cPickle as pickle
except ImportError:
    import pickle
from decimal import Decimal
from redis import WatchError
from .stores import Store
from .serials import SerialsManager, Serial, PeriodSerial, PERIOD_DAILY

STATE_NEW, STATE_DIRTY, STATE_UNCHANGED = 'new', 'dirty', 'unchanged'

MAX_EXPIRE_SECONDS = 946656000  # timestamp of Jan. 1st, 2000
                                # objects' __expired__ large than this
                                # will be treated as a timestamp,
                                # otherwise it will be treated as seconds


class Attr(object):
    """Attributes for Hash Objects
    
    The base Attr using pickle to load and dump data.
    
    Attributes:
    
        name:  attribute name of object in Python
        field: field name of hash in Redis, by default its same as name
        default: default value
        required: Does this attr required? A required attr must not be None.
                  Note that not all Attr subclasses has this options.
    """
    
    def __init__(self, field=None, default=None, required=False):
        self.field = field   # hash field in Redis
        self.name = None     # attribute name in Python
        self.default = default
        self.required = bool(required)
        self.as_key = False

    def get_value(self, obj):
        storage = obj.__ohm_storage__
        if self.name not in storage:
            self.set_value(obj, self.default)
        return storage[self.name]

    def set_value(self, obj, value):
        storage = obj.__ohm_storage__
        if self.as_key and obj.__fixed_keys__ and obj.__ohm_state__!=STATE_NEW:
            raise RuntimeError('cannot change a key attr while '
                               '"__fixed_keys__" is True')
        original = storage.get(self.name, None)
        if obj.__ohm_state__==STATE_UNCHANGED:
            if value!=original or type(value)!=type(original):
                obj.__ohm_state__ = STATE_DIRTY
        storage[self.name] = value

    def validate(self, value):
        if self.required and value == None:
            raise ValueError('"%s" is required but it is None' % self.name)

    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        data = pickle.dumps(value)
        return data

    def loads(self, obj, data):
        value = pickle.loads(data)
        self.set_value(obj, value)


class EnumAttr(Attr):
    def __init__(self, field=None, default=None, required=False,
                 values=None):
        if type(values) not in [tuple, list, set]:
            raise TypeError('"values" should be tuple, list or set')
        values = list(values)
        Attr.__init__(self, field, default, required)
        if not self.required:
            values.append(None)
        if self.default!=None and self.default not in values:
            raise ValueError('"default" neither in values nor None')
        self.values = set(values)

    def validate(self, value):
        if value not in self.values:
            raise ValueError('"%r" is invalid' % value)


class NumberAttr(Attr):
    def __init__(self, field=None, default=Decimal('0'),
                 min_value=None, max_value=None):
        Attr.__init__(self, field, default, True)
        self.min_value = min_value
        self.max_value = max_value
    
    def validate(self, value):
        if self.min_value != None and value < self.min_value:
            raise ValueError('value is too small')
        if self.max_value != None and value > self.max_value:
            raise ValueError('value is too large')
    
    def set_value(self, obj, value):
        value = Decimal(value)
        Attr.set_value(self, obj, value)

    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        return str(value)

    def loads(self, obj, data):
        self.set_value(obj, data)


class IntAttr(NumberAttr):
    def __init__(self, field=None, default=0, as_key=False,
                 min_value=None, max_value=None):
        self.as_key = bool(as_key)
        NumberAttr.__init__(self, field, default, min_value, max_value)

    def set_value(self, obj, value):
        value = int(value)
        self.validate(value)
        Attr.set_value(self, obj, value)


class TimestampAttr(IntAttr):
    def __init__(self, field=None, default=None):
        IntAttr.__init__(self, field, default, False, 0, 2**32-1)

    def set_value(self, obj, value):
        if value==None:
            value = int(time.time())
        IntAttr.set_value(self, obj, value)    


class SerialAttr(IntAttr):
    def __init__(self, field=None, default=None, as_key=False):
        IntAttr.__init__(self, field, default, as_key, None, None)

    def set_value(self, obj, value):
        if value!=None:
            IntAttr.set_value(self, obj, value)
        else:
            Attr.set_value(self, obj, value)
    
    def dumps(self, obj):
        value = self.get_value(obj)
        if value==None:
            classname = obj.__class__.__name__
            name = '%s.%s' % (classname, self.field)
            serials = obj.manager.serials
            if not serials.has_serial(name):
                serial = Serial(name, pattern=u'%d')
                serials.add_serial(serial)
            value = serials.incr(name)
            self.set_value(obj, value)
        return str(value)


class FloatAttr(NumberAttr):
    def __init__(self, field=None, default=0.0,
                 min_value=None, max_value=None):
        NumberAttr.__init__(self, field, default, min_value, max_value)

    def set_value(self, obj, value):
        # NOTE: Just use str() may cause a problem:
        #
        # >>> 1234567890.1234567890
        # 1234567890.1234567
        # >>> str(1234567890.1234567890)
        # '1234567890.12'
        # >>> float(_)
        # 1234567890.12
        #
        # So we use decimal.Decimal for converting:
        #
        # >>> 1234567890.1234567890
        # 1234567890.1234567
        # >>> decimal.Decimal(1234567890.1234567890)
        # Decimal('1234567890.1234567165374755859375')
        # >>> float(_)
        # 1234567890.1234567
        # >>> str(decimal.Decimal(1234567890.1234567890))
        # '1234567890.1234567165374755859375'
        # >>> float(_)
        # 1234567890.1234567
        value = float(Decimal(value))
        Attr.set_value(self, obj, value)

    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        # without help of Decimal, dump a float then load it
        # may result a slightly different value
        data = str(Decimal(value))
        return data


class BooleanAttr(Attr):
    def __init__(self, field=None, default=None, required=False):
        Attr.__init__(self, field, default, required)

    def set_value(self, obj, value):
        value = str(value).strip().lower()
        if value in ['true', 'yes', 'on', 't', 'y', '1']:
            value = True
        elif value in ['false', 'no', 'off', 'f', 'n', '0']:
            value = False
        elif value in ['none', 'null', 'nil', 'n/a', 'na', '~', '']:
            value = None
        else:
            raise TypeError('unsupport value BooleanAttr value type')
        Attr.set_value(self, obj, value)

    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        if value:
            data = 'y'
        elif value == False:
            data = 'n'
        else:
            data = '~'
        return data

    def loads(self, obj, data):
        self.set_value(obj, data)


class StringAttr(Attr):
    def __init__(self, field=None, default='', as_key=False,
                 min_length=0, max_length=None, encoding='utf8'):
        Attr.__init__(self, field, default, True)
        self.as_key = bool(as_key)
        self.min_length = min_length
        self.max_length = max_length
        self.encoding = encoding
    
    def validate(self, value):
        if len(value) < self.min_length:
            raise ValueError('value is too short')
        if self.max_length != None and len(value) > self.max_length:
            raise ValueError('value is too long')
    
    def set_value(self, obj, value):
        if type(value) not in [int, long, str, unicode]:
            raise TypeError('unsupport StringAttr value type')
        if type(value)==unicode:
            value = value.encode(self.encoding)
        else:
            value = str(value)
        Attr.set_value(self, obj, value)

    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        return value

    def loads(self, obj, data):
        self.set_value(obj, data)


class PeriodSerialAttr(StringAttr):
    def __init__(self, field=None, default=None, as_key=False,
                 min_length=0, max_length=None, encoding='utf8',
                 period=PERIOD_DAILY, pattern=u'%(moment)s%(number)08d',
                 moment_format=None):
        StringAttr.__init__(self, field, default, as_key, None, None, encoding)
        self.period = str(period)
        self.pattern = unicode(pattern)
        self.moment_format = unicode(moment_format) if moment_format else None

    def set_value(self, obj, value):
        if value!=None:
            StringAttr.set_value(self, obj, value)
        else:
            Attr.set_value(self, obj, value)
    
    def dumps(self, obj):
        value = self.get_value(obj)
        if value==None:
            classname = obj.__class__.__name__
            name = '%s.%s' % (classname, self.field)
            serials = obj.manager.serials
            if not serials.has_serial(name):
                serial = PeriodSerial(name, self.period, pattern=self.pattern,
                                      moment_format=self.moment_format)
                serials.add_serial(serial)
            value = serials.incr(name)
            self.set_value(obj, value)
        return str(value)


class UnicodeAttr(StringAttr):

    def set_value(self, obj, value):
        if type(value) not in [int, long, str, unicode]:
            raise TypeError('unsupport UnicodeAttr value type')
        if type(value)==str:
            value = value.decode(self.encoding)
        else:
            value = unicode(value)
        Attr.set_value(self, obj, value)
    
    def dumps(self, obj):
        value = self.get_value(obj)
        self.validate(value)
        return value.encode(self.encoding)


class HashMeta(type):
    def __new__(class_, name, bases, dict_):
        meta = dict(
            attrs=dict(),   # attr name => instance of Attr
            names=dict(),   # redis field name => ohm attr name
            keys=list(),    # list of attr names which has been marked as_key
        )
        for base_class in bases:
            if hasattr(base_class, '__ohm_meta__'):
                super_meta = base_class.__ohm_meta__
                meta['attrs'].update(super_meta['attrs'])
                meta['names'].update(super_meta['names'])
                meta['keys'].extend(super_meta['keys'])
        for itemname in dict_:
            attr = dict_[itemname]
            if isinstance(attr, Attr):
                attr.name = itemname
                if not attr.field:
                    attr.field = itemname
                meta['attrs'][attr.name] = attr
                meta['names'][attr.field] = attr.name
                if attr.as_key:
                    meta['keys'].append(attr.field)
        for itemname in meta['attrs']:
            attr = meta['attrs'][itemname]
            dict_[itemname] = property(attr.get_value, attr.set_value)
        dict_['__ohm_meta__'] = meta
        return type.__new__(class_, name, bases, dict_)


class Hash(object):
    """Hash
    """
    __metaclass__ = HashMeta
    __expired__ = 0
    __fixed_keys__ = False
    __ensure_new__ = False
    
    def __new__(class_, *args, **kwargs):
        obj = super(Hash, class_).__new__(class_)
        obj.__ohm_storage__ = dict()
        obj.__ohm_manager__ = None
        obj.__ohm_state__ = STATE_NEW
        meta = class_.__ohm_meta__
        if not meta['keys']:
            field = '_id'
            attr = SerialAttr(field, as_key=True)
            attr.name = field
            meta['attrs'][attr.name] = attr
            meta['names'][attr.field] = attr.name
            meta['keys'].append(attr.field)
            setattr(class_, attr.name, property(attr.get_value, attr.set_value))
        meta['keys'].sort()
        return obj

    @classmethod
    def _get_attrs(cls):
        meta = cls.__ohm_meta__
        return meta['attrs'].copy()
    
    @classmethod
    def _get_attr_by_name(cls, name):
        return cls._get_attrs()[name]
    
    @classmethod
    def _get_attr_by_filed(cls, field):
        meta = cls.__ohm_meta__
        return cls._get_attr_by_name(meta['names'][field])
    
    @classmethod
    def _get_key_prefix(cls):
        return getattr(cls, '__key_prefix__', cls.__name__)
    
    @classmethod
    def _get_key_sep(cls):
        return getattr(cls, '__key_sep__', ':')
    
    @classmethod
    def _get_key_fields(cls):
        meta = cls.__ohm_meta__
        return meta['keys'][:]

    @classmethod
    def _get_key_pattern(cls):
        meta = cls.__ohm_meta__
        prefix = cls._get_key_prefix()
        sep = cls._get_key_sep()
        fields = cls._get_key_fields()
        if len(fields)==1:
            field = fields[0]
            pattern = '%s%s%%(%s)s' % (prefix, sep, field)
        else:
            pattern = prefix
            for field in fields:
                pattern += '%s%s%s%%(%s)s' % (sep, field, sep, field)
        return pattern

    def _get_key(self, data):
        pattern = self._get_key_pattern()
        return pattern % data
    
    def attach(self, manager):
        if not self.__ohm_manager__:
            self.__ohm_manager__ = manager
        elif self.__ohm_manager__ != manager:
            raise RuntimeError('this object has attached to a manager')
    
    def detach(self):
        self.__ohm_manager__ = None
    
    @property
    def manager(self):
        return self.__ohm_manager__
    
    def save(self, manager=None):
        manager = manager or self.manager
        if not manager:
            raise RuntimeError('this object has not been attached to a manager')
        manager.save(self)

    def _dump(self, manager):
        self.attach(manager)
        data = dict()
        attrs = self._get_attrs()
        for name in attrs:
            attr = attrs[name]
            field = attr.field
            value = attr.dumps(self)
            data[field] = value
        return data
    
    def _load(self, data, manager):
        self.attach(manager)
        for field in data:
            value = data[field]
            try:
                attr = self._get_attr_by_filed(field)
            except KeyError:
                attr = None
            if attr:
                attr.loads(self, value)
    
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        attrs = self._get_attrs()
        for name in attrs:
            if getattr(self, name) != getattr(other, name, None):
                return False
        return True
    
    def __ne__(self, other):
        return not self.__eq__(other)


class Manager(Store):
    """Object Hash Mapper Manager
    """

    def __init__(self, store, serials_key='<OHM-SERIALS>'):
        Store.__init__(self, store)
        self.serials = SerialsManager(store, str(serials_key))

    def save(self, obj):
        if not isinstance(obj, Hash):
            raise TypeError('obj should be an instance of Hash')
        # get data and key
        state = obj.__ohm_state__
        expired = int(obj.__expired__)
        data = obj._dump(self)
        key = obj._get_key(data)
        # does "expire" or "expireat" command required?
        expire_required = False
        if state==STATE_NEW and expired!=0:
            expire_required = True
        # save it and expire it if needed
        if state==STATE_NEW and bool(obj.__ensure_new__):
            # if __ensure_new__ is True, work in a transaction
            with self.store.pipeline() as trans:
                try:
                    trans.watch(key)  # watch a nonexistent key is OK
                    if trans.exists(key):
                        raise RuntimeError
                    trans.multi()
                    success = trans.hmset(key, data)
                    # may raise redis.WatchError below
                    trans.execute()
                except (RuntimeError, WatchError):
                    raise RuntimeError('key "%s" exists' % key)
        else:
            # may raise redis.ResponseError below
            success = self.store.hmset(key, data)
        if success:
            obj.__ohm_state__ = STATE_UNCHANGED
            if expire_required:
                if expired > MAX_EXPIRE_SECONDS:
                    self.store.expireat(key, expired)
                else:
                    self.store.expire(key, expired)
        return success

    def load(self, class_, key):
        if not issubclass(class_, Hash):
            raise ValueError('argument "class_" should be subclass of Hash')
        data = self.store.hgetall(key)
        if not data:
            return None
        obj = class_()
        obj._load(data, self)
        obj.__ohm_state__ = STATE_UNCHANGED
        return obj
    
    def find(self, class_, *pattern, **kwargs):
        if not issubclass(class_, Hash):
            raise ValueError('argument "class_" should be subclass of Hash')
        if pattern:
            # has pattern, search for "Object:<pattern>"
            # mapping is ignore
            pattern = pattern[0]
            prefix = class_._get_key_prefix()
            sep = class_._get_key_sep()
            pattern = '%s%s%s' % (prefix, sep, pattern)
        elif not kwargs:
            # neither pattern not mapping, search for "Object:*"
            prefix = class_._get_key_prefix()
            sep = class_._get_key_sep()
            pattern = '%s%s*' % (prefix, sep)
        else:
            # only the mapping is given, search for
            # "Object:key1:value1:key2:value2:not-given-key:*"
            key_pattern = class_._get_key_pattern()
            key_fields = class_._get_key_fields()
            for field in key_fields:
                if field not in kwargs:
                    kwargs[field] = '*'
            pattern = key_pattern % kwargs
        keys = self.store.keys(pattern)
        return keys



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

import time
from datetime import datetime
import redis
import aloharedis
assert aloharedis.__version__ == '0.0.3'
from aloharedis.serials import *

from nose.tools import *

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! IMPORTANT: We use redis db=7 for these tests   !!!
# !!! and will DELETE ALL DATA from it after tested. !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DB = 7

def setup_module():
    global redis_client
    redis_client = redis.StrictRedis(db=DB)


def teardown_module():
    redis_client.flushdb()


def test_serial():
    store = SerialsManager(redis_client, '<serials>',
        Serial('sn', 1, 2, u'SN:%010d')
    )
    eq_(u'SN:0000000003', store.incr('sn'))     # 1+2
    eq_(u'SN:0000000005', store.incr('sn'))     # 3+2
    eq_(u'SN:0000000008', store.incr('sn', 3))  # 5+3
    eq_(u'SN:0000000010', store.incr('sn', 0))  # 8+2


def test_period():
    fmt = u'%y/%m/%d'
    store = SerialsManager(redis_client, '<serials>',
        PeriodSerial('bills', PERIOD_DAILY,
                     pattern=u'B/%(moment)s/%(number)07d',
                     moment_format=fmt)
    )
    moment = datetime.now().strftime(fmt)
    eq_(u'B/'+moment+u'/0000001', store.incr('bills'))
    eq_(u'B/'+moment+u'/0000002', store.incr('bills'))


def test_outdated():
    period = '%Y%m%d%H%M%S'
    store = SerialsManager(redis_client, '<secondly-serials>',
        PeriodSerial('secondly', period, pattern=u'%(number)d')
    )
    eq_('1', store.incr('secondly'))
    time.sleep(1.0)
    eq_('1', store.incr('secondly'))
    time.sleep(1.0)
    eq_('1', store.incr('secondly'))
    ok_(len(store.keys)==3)
    time.sleep(1.0)
    store.clean()
    ok_(len(store.keys)==0)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

from distutils.core import setup
from aloharedis import __version__

with open('README.rst', 'r') as f:
    long_description = f.read()


setup(
    name = "aloharedis",
    version = __version__,
    packages = ["aloharedis"],
    requires = ["redis"],
    author = "Xue Can",
    author_email = "xuecan@gmail.com",
    url = "https://bitbucket.org/xuecan/aloharedis",
    keywords = ["redis", "lock", "cache", "ohm", "object hash mapper"],
    description = "Aloharedis is a set of tools for working with redis-py.",
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    long_description = long_description
)



# Copyright (C) 2012 Xue Can <xuecan@gmail.com> and contributors.
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license

import time
import uuid
from datetime import datetime
from random import random
import redis
import aloharedis
assert aloharedis.__version__ == '0.0.3'
from aloharedis.caches import *

from nose.tools import *

# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# !!! IMPORTANT: We use redis db=7 for these tests   !!!
# !!! and will DELETE ALL DATA from it after tested. !!!
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
DB = 7

def setup_module():
    global redis_client
    redis_client = redis.StrictRedis(db=DB)


def teardown_module():
    redis_client.flushdb()


def test1():
    c = CacheManager(redis_client, default_timeout=1)
    eq_(None, c.get('a'))
    data = uuid.uuid4()
    c.set('a', data)
    eq_(data, c.get('a'))
    c.delete('a')
    eq_(None, c.get('a'))
    c.set('a', data)
    time.sleep(2)
    eq_(None, c.get('a'))
    c.set('a', 1)
    c.set('b', 2)
    c.set('c', 3)
    eq_([1,2,3,None], c.get_many('a', 'b', 'c', 'd'))
    eq_(dict(a=1,b=2,c=3,d=None), c.get_dict('a', 'b', 'c', 'd'))
    c.add('a', 100)
    eq_(1, c.get('a'))
    c.set_many(dict(x=1,y=2,z=3))
    eq_([1,2,3], c.get_many('x', 'y', 'z'))
    eq_(2, c.inc('x'))
    eq_(4, c.inc('x', 2))
    eq_(2, c.dec('x', 2))
    eq_(1, c.dec('x'))
    c.delete_many('x','y','z')
    eq_([None,None,None], c.get_many('x', 'y', 'z'))
    c.set('a', 'x', 10)
    time.sleep(2)
    eq_(['x', None], c.get_many('a', 'b'))
    eq_(['<CACHE>a'], c.keys())
    c.clear()
    eq_([], c.keys())


def test_ts_as_timeout():
    c = CacheManager(redis_client)
    n = int(time.time())
    c.set('a', 'xxx', n+2)
    eq_('xxx', c.get('a'))
    time.sleep(3)
    eq_(None, c.get('a'))
    